#include<pic18.h>
#include"delay.h"
#include "serial.h"


#define MAXDATA_LEN 1280
#define MAXTRAME_LEN 48

#define uchar unsigned char
#define vub unsigned char
#define vsb signed char

#define vuw unsigned short int
#define vsw signed short int

#define vud unsigned long int
#define vsd signed long int

vub PREC_BEFORE=4;
vub PREC_AFTER=4;
vub sound_on=1;

vud xtimer;
vsb dtimer;

void interrupt ISR()
{
    if((TMR0IF)&&(TMR0IE))
	{
		xtimer+=262144;
		if (xtimer>1000000)
		{
			xtimer-=1000000;
			dtimer++;
		}
		TMR0IF=0;
	}
}

void starttimer()
{
	xtimer=0;
	dtimer=0;
	GIEH=1;	// suppression des interruptions
	GIEL=1;
	IPEN=0;
	TMR0H=0;
	TMR0L=0;
	
	T0CON=0x81;	// lancement du timer 0 ; essayer 0x87
	TMR0IE=1;
	TMR0IF=0;
}

void halttimer()
{
	GIEH=0;	// suppression des interruptions
	GIEL=0;
	IPEN=0;
	TMR0IE=0;
}

char time0()
{
	char a;
	a=TMR0L;
	return TMR0H;
}

void vprint(const char* p,uchar ln)
{
	char c;
	while((c = *p++) != 0)
	{
		putch(c);
	}
	while(ln-->0)
	{
		putch('\r');
		putch('\n');
	}
}

char webHexChar(char val)
{
	if (val<10) return '0'+val;
	return 'A'+val-10;
}

void vprinth(char v,uchar ln)
{
	putch(webHexChar((v>>4)&15));
	putch(webHexChar(v&15));
	while(ln-->0)
	{
		putch('\r');
		putch('\n');
	}
}
void vprinthw(vuw v,uchar ln)
{
	vprinth(v>>8,0);
	vprinth(v,ln);
}

#define SOUND RA5
#define TAPIS1 RD5
//#define TAPIS1 RD4
#define TAPIS2 RC2
//#define TAPIS2 RD5
#define TAPIS4 RD6
#define TAPIS8 RD7
#define TAPIS16 RD4
//#define TAPIS16 RC2

#define LEDGREEN RB5
#define LEDRED RB4
#define DIGIT6 RB3
#define DIGIT5 RB2
#define DIGIT4 RB1
#define DIGIT3 RB0
#define DIGIT2 RE2
#define DIGIT1 RE1
#define DIGIT0 RE0

// gestion de l'affichage

const vub digit[]=
{0x77,0x24,0x5d,0x6d,0x2e,0x6b,0x7b,0x25,
 0x7f,0x6f,0x3f,0x7a,0x53,0x7c,0x5b,0x1b};

vub currentdigit=0;
vub countred=0;
vub countgreen=0;

vub ledfromstep(vub step)
{
	vub i=0;
	if (step&1) i+=0x12;
	if (step&2) i+=0x40;
	if (step&4) i+=0x01;
	if (step&8) i+=0x24;
	if (step&16) i+=0x20;
	return i;
}

void drawleds()
{
	vub digit=currentdigit;

	DIGIT0=(digit&1)?0:1;
	DIGIT1=(digit&2)?0:1;
	DIGIT2=(digit&4)?0:1;
	DIGIT3=(digit&8)?0:1;
	DIGIT4=(digit&16)?0:1;
	DIGIT5=(digit&32)?0:1;
	DIGIT6=(digit&64)?0:1;

	LEDRED=(countred)?0:1;
	LEDGREEN=(countgreen)?0:1;
	if ((countred)&&(countred!=255)) countred--;
	if ((countgreen)&&(countgreen!=255)) countgreen--;
}

void setdigithex(vub val)
{
	currentdigit=digit[val&15];
}

void setdigitstep(vub step)
{
	currentdigit=ledfromstep(step);
}

void setdigit(vub val)
{
	currentdigit=val;
}

void setred(vub val)
{
	countred=val;
}

void setgreen(vub val)
{
	countgreen=val;
}

// gestion du tapis
vub counttapis[5];

void checkstep()
{
	vub rstep,i,j;
	rstep=(TAPIS1)+(TAPIS2<<1)+(TAPIS4<<2)+(TAPIS8<<3)+(TAPIS16<<4);
	j=1;
	for(i=0;i<5;i++)
	{
		counttapis[i]=(rstep&j)?((counttapis[i]!=255)?counttapis[i]+1:counttapis[i]):0;
		j<<=1;
	}
}

vub getstep(vub lim)
{
	vub step,i,j,k;
	step=0;
	j=1;
	for(i=0;i<5;i++)
	{
		k=counttapis[i];
		step+=((k)&&(k<=lim))?j:0;
		j<<=1;
	}
	return step;
}

vub getstepfutur(vub lim,vub futur)
{
	vub step,i,j,k;
	step=0;
	j=1;
	for(i=0;i<5;i++)
	{
		k=counttapis[i];
		step+=((k)&&(k+futur<=lim))?j:0;
		j<<=1;
	}
	return step;
}

#define waittransition \
	asm("nop");	\
	asm("nop");	\
	asm("nop");	\
	asm("nop");	\
	asm("nop");	\
	asm("nop");	\


#define TIMEOUT_MAX_SYNCH 8
#define TIMEOUT_MAX_COLUMN 10
#define TIMEOUT_MIN_SYNCV 20
vub res;
vub flagres;
void getlinedata()
{
	vub i,r,current,timeout;

	flagres=1;
	res=0;

	r=0;
	i=8;
	current=1;

// attente dbut de ligne
	do
	{
		timeout=1;
		while((PORTD&1));
		while(!(PORTD&1)) if (timeout) timeout++;
	} while((timeout)&&(timeout<TIMEOUT_MAX_SYNCH));


//	while(!(PORTD&1));	// on attend la fin du noir
	while(i)
	{
		timeout=0;
		r<<=1;
		if (current)
		{
			while((PORTD&1)) timeout++;
			if (timeout>TIMEOUT_MAX_COLUMN) return;
			waittransition
			if (PORTD&1) r++;
			else current--;
		}
		else
		{
			while(!(PORTD&1)) timeout++;
			if (timeout>TIMEOUT_MAX_COLUMN) return;
			waittransition
			if (!(PORTD&1)) r++;
			else current++;
		}
		i--;
	}
	flagres=0;
	res=r;
}

const char* crypt="amberMind";

vub buftrame[MAXTRAME_LEN];
vub buftramelen;

void getpicdata()
{
	vub timeout;

	vub i,x,y,n,icr,ok;

// attente dbut cran
	do
	{
		timeout=1;
		while((PORTD&1));
		while(!(PORTD&1)) if (timeout) timeout++;
	} while((timeout)&&(timeout<TIMEOUT_MIN_SYNCV));

	i=0;
	x=0;
	n=0;
	icr=0;
	ok=1;
	while(ok)
	{
		getlinedata();
		y=res;
		if ((y==x)&&(flagres==0))
		{
			n++;
			if (n==2)//((n==2)||(n==7))
			{
				res-=crypt[icr];
				icr++; if (icr>=9) icr=0;
				buftrame[i]=res;
				if (i)
				{
					if(!res) ok=0;
				}
				else if (res) ok=i=0;

				i++;
				if (i>=MAXTRAME_LEN) ok=0;
			}
		}
		else
		{
			x=y;
			n=0;
		}
	}
	buftramelen=i;
}

vub waitendmagic()
{
	vub i,timeout;

// attente dbut cran
	do
	{
		timeout=1;
		while((PORTD&1));
		while(!(PORTD&1)) if (timeout) timeout++;
	} while((timeout)&&(timeout<TIMEOUT_MIN_SYNCV));

	for(i=0;i<10;i++)
	{
		getlinedata();
		if ((flagres==0)&&(res==crypt[0])) return 0;
	}
	return 1;
}

void gettramedata(vub oldtrame)
{
	vub x,y,n;

	x=0;
	n=0;
	while(1)
	{
		getpicdata();
		if (buftramelen)
		{
			y=buftrame[1];
			if ((y!=oldtrame)&&(y==x))
			{
				n++;
				if (n>1) return;
			}
			else n=0;
			x=y;
		}
	}
}



/*
void displaytrame()
{
	int x;
	for(x=0;x<buftramelen;x++)
	{
		vprinth(x,0);vprint("->",0);vprinth(buftrame[x],1);
	}
}
*/

vub bufdata[MAXDATA_LEN];
vuw bufdatalen;

const vub tabstep[]=
{0,1,2,4,8,16,33,5,
 9,17,6,10,18,12,20,24};

void sound_beep()
{
	vub i;
	for(i=0;i<20;i++)
	{
		SOUND=1;
		DelayUs(250);
		SOUND=0;
		DelayUs(250);
	}
//	vprint("beep!",1);
}
void sound_bop()
{
	vub i;
	for(i=0;i<10;i++)
	{
		SOUND=1;
		DelayMs(2);
		SOUND=0;
		DelayMs(2);
	}
//	vprint("bop!",1);
}

void dvdgame()
{
	vsd ind,i;
	vub	delay;
	vub	step;
	vsb oldtimer,t0,okdone;
	vuw goodstep;
	vuw nbstep;
	vub delaylong,steplong;
	vub nbreak;

	vprinthw(bufdatalen,1);
	ind=0;	// on va supprimer les codes "nop" (0x01)
	for(i=0;i<bufdatalen;i++)
	{
		if (bufdata[i]!=1)
		{
			if (ind!=i) bufdata[ind]=bufdata[i];
			ind++;
		}
	}
	bufdatalen=ind;
	vprinthw(bufdatalen,1);

//	for(i=0;i<bufdatalen;i++) vprinth(bufdata[i],1);

	while(!waitendmagic());
	setdigit(0);
	drawleds();

	ind=0;
	goodstep=0;
	nbstep=0;
	starttimer();
//	while(dtimer<50);
	while(dtimer<48+PREC_AFTER);
//	while(dtimer<50+PREC_AFTER);
	dtimer=0;
	oldtimer=0;
	t0=0;
	okdone=0;
	nbreak=0;

	while(ind<bufdatalen)
	{
		delay=bufdata[ind++];
		step=tabstep[delay&15];
		delay=(delay>>4)&15;
		if (step)
		{
			delaylong=delay;
			steplong=step;
		}
		else
		{
			delaylong=bufdata[ind];
			steplong=tabstep[delaylong&15];
			delaylong=(delaylong>>4)&15;
		}

		do
		{
			i=dtimer;
			if (i!=oldtimer)
			{
				checkstep();
//				setdigitstep(getstep(PREC_BEFORE+PREC_AFTER));
				if ((!okdone)&&(steplong)&&((steplong&getstepfutur(PREC_BEFORE+PREC_AFTER,t0+delaylong-i))==steplong))
				{
					setgreen(5);
					goodstep++;
					okdone=1;
					if (sound_on) sound_beep();
				}
//				vprinth(getstep(PREC_BEFORE+PREC_AFTER),1);
				drawleds();
				oldtimer++;
				if (!waitendmagic()) nbreak++;
				else nbreak=0;
				if (nbreak>=10)
				{
					vprint("break",1);
					setdigithex(0xb);
					drawleds();
					halttimer();
	 				return;
				}
			}
			i-=t0;
			i&=255;
		}
		while(i<delay);
		t0+=delay;
		if (step)
		{
			nbstep++;
			if ((!okdone)&&((step&getstep(PREC_BEFORE+PREC_AFTER))!=step))
			{
				setred(5);
				if (sound_on) sound_bop();
			}
			setdigithex(goodstep?(goodstep*10-1)/nbstep:0);
			vprinth(step,1);
			okdone=0;
		}
	}
	halttimer();
}


void dvdtronic()
{
	vub oldop,op,i,ok;


	vprint("start DVDTRONIC",1);
	setdigit(0);
	drawleds();
//	getpicdata();
//	displaytrame();

	oldop=0;
	ok=1;
	while(ok)
	{
		gettramedata(oldop);
//		displaytrame();
		op=buftrame[1];
		oldop=op;
		op>>=4;
		if (op==1)
		{
			// DVD RESET
		}
		else if (op==2)
		{
			// DVD ECHO
		}
		else if ((op==3)||(op==4))
		{
			setdigithex(0xd);
			drawleds();
			// DVD DATA
			if (op==3) bufdatalen=0;
			if (bufdatalen+buftramelen>=MAXDATA_LEN)
			{
				// ERROR !!! trame trop longue
				return;
			}
			for(i=2;i<buftramelen-1;i++)bufdata[bufdatalen++]=buftrame[i];
		}
		else if (op==5)
		{
			setdigithex(0);
			drawleds();
			// DVD START
			dvdgame();
		}
		else if (op==6)
		{
//			setdigithex(6);
			drawleds();
			if (buftrame[2])
			{
				PREC_BEFORE=2;
				PREC_AFTER=2;
				sound_beep();
			}
			else
			{
				PREC_BEFORE=4;
				PREC_AFTER=4;
				sound_bop();
			}
			vprint("level ",0);vprinth(PREC_BEFORE,1);
		}
		else if (op==7)
		{
//			setdigithex(7);
			drawleds();
			if (buftrame[2])
			{
				sound_on=1;
				sound_beep();
			}
			else
			{
				sound_on=0;
				sound_bop();
			}
			vprint("sound ",0);vprinth(sound_on,1);
		}
	}
	vprint("\r\nok",1);
}

	
void main()
{
	char x;
//	char tc;
	vub step,oldstep;
	char i;

	TRISB=0x00;	// tout en sortie
	TRISA4=0;	// RA4 en entre
	TRISA5=0;	// RA5 en sortie

	PORTD=0;
	LATD=0;
	TRISD=0xff;	// RD en entre
	TRISC|=4;
	ADCON1=6;
	TRISE=0;

//	TRISD1=1;	// RD1 en entre
	PORTB=0x00; // leds teintes

//	starttimer();
	halttimer();

	init_comms();

	vprint("\r\nSTART\r\n",0);
  

	oldstep=0;
	for(i=0;i<5;i++)counttapis[i]=0;
	x=0;
	while(1)
	{
		checkstep();
		step=getstep(5);
		setdigitstep(step);
		if (step!=oldstep)
		{
			vprinth(step,1);
			if (step==1) sound_beep();
			if (step==8) sound_bop();
			oldstep=step;
		}
		setgreen(x&8);
		setred(1);
		drawleds();
		x++;
		if (!waitendmagic()) dvdtronic();
		DelayMs(50);
	}
}
